#!/usr/bin/env python3
"""Entrypoint for UI."""
import os
from .backend import detect_display
os.environ["GDK_BACKEND"] = detect_display()

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib

from . import settings
from .css import get_css_provider
from .file import init_tlp_file_config
from .mainui import create_main_box, store_window_size, window_key_events, close_main_window, reset_scroll_position

Gtk.init()
Gtk.StyleContext.add_provider_for_screen(
    Gdk.Screen.get_default(),
    get_css_provider(),
    Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
)

# Set window properties
GLib.set_prgname('tlpui')
Gdk.set_program_class('Tlp-UI')

# Apply custom scalable icons to icon theme
Gtk.IconTheme().get_default().append_search_path(settings.icondir + 'themeable')


def main() -> None:
    """Initiate main window with all sub elements."""
    # init configuration settings
    init_tlp_file_config()

    # init application window
    window = Gtk.Window()
    window.set_title('Tlp-UI')
    window.set_icon_name('tlpui')
    window.set_default_size(settings.userconfig.windowxsize, settings.userconfig.windowysize)
    window.add(create_main_box(window))
    window.connect('check-resize', store_window_size)
    window.connect('delete-event', close_main_window)
    window.connect('key-press-event', window_key_events)
    window.show_all()
    reset_scroll_position()
    Gtk.main()


if __name__ == '__main__':
    main()



"""Disk view UI widget."""

from gi.repository import Gtk
from .. import settings


def create_view(configname: str) -> Gtk.Box:
    """Create disk view."""
    tlpobject = settings.tlpconfig[configname]
    label = Gtk.Label(tlpobject.get_value())
    label.set_width_chars(len(tlpobject.get_value()) + 5)
    return label



"""TLP config."""

from enum import Enum
from .uihelper import StateImage


class ConfType(Enum):
    """Config type enum class."""

    DEFAULT = 1
    DROPIN = 2
    USER = 3
    ERR = 4


class TlpConfig:
    """TLP config class."""

    def __init__(self, enabled: bool, name: str, value: str, conftype: ConfType, confpath=""):
        """Init TLP config class parameters."""
        self.enabled = enabled
        self.enabledstore = enabled
        self.name = name
        self.value = value
        self.valuestore = value
        self.conftype = conftype
        self.confpath = confpath
        self.stateimage = None      # type: StateImage

    def get_name(self) -> str:
        """Get config name."""
        return self.name

    def get_conf_type(self) -> ConfType:
        """Get config type."""
        return self.conftype

    def get_conf_path(self) -> str:
        """Get config path."""
        return self.confpath

    def get_value(self) -> str:
        """Get config value."""
        return self.value

    def set_value(self, newvalue: str):
        """Get config value."""
        self.value = newvalue
        self.refresh_state_image()

    def set_enabled(self, newstate: bool):
        """Set config enabled."""
        self.enabled = newstate
        self.refresh_state_image()

    def is_enabled(self) -> bool:
        """Get config enabled."""
        return self.enabled

    def add_state_image(self, newstateimage: StateImage):
        """Add image with state."""
        self.stateimage = newstateimage
        self.refresh_state_image()

    def refresh_state_image(self):
        """Refresh state image."""
        self.stateimage.refresh(self.value, self.valuestore, self.enabled, self.enabledstore)



"""Switch widget."""

from gi.repository import Gtk
from .. import settings


def create_state_switch(configname: str, values: str) -> Gtk.Switch:
    """Create switch."""
    switch = Gtk.Switch()

    toggleitems = values.split(',')
    falseitem = toggleitems[0]
    trueitem = toggleitems[1]

    if settings.tlpconfig[configname].get_value() == trueitem:
        switch.set_active(True)
    else:
        switch.set_active(False)

    switch.connect('notify::active', change_switch_state, configname, falseitem, trueitem)
    return switch


def change_switch_state(self: Gtk.Switch, notify, configname: str, falseitem: str, trueitem: str):
    """Process and store state change."""
    tlpobject = settings.tlpconfig[configname]

    if self.get_active():
        tlpobject.set_value(trueitem)
    else:
        tlpobject.set_value(falseitem)



"""Entry UI widget."""

from gi.repository import Gtk
from .. import settings


def create_entry(configname: str) -> Gtk.Entry:
    """Create entry widget."""
    configvalue = settings.tlpconfig[configname].get_value()
    configvaluelength = len(configvalue)
    if configvaluelength >= 70:
        configvaluelength = 70

    entry = Gtk.Entry()
    entry.set_text(configvalue)
    entry.set_width_chars(configvaluelength + 5)
    entry.connect('changed', change_entry_text, configname)
    return entry


def change_entry_text(self: Gtk.Entry, configname: str):
    """Store state change."""
    settings.tlpconfig[configname].set_value(self.get_text())



"""This module provides error helper functions for the UI."""

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk


def show_dialog(error_message) -> None:
    """Show error dialog."""
    dialog = Gtk.MessageDialog()
    dialog.set_default_size(150, 100)
    dialog.add_button(Gtk.STOCK_OK, Gtk.ResponseType.OK)
    dialog.format_secondary_markup(error_message)
    dialog.run()
    dialog.destroy()



"""This module provides helper functions for application settings."""

import configparser
import re
import sys
from os import getenv
from shutil import which
from subprocess import check_output, CalledProcessError
from pathlib import Path
from . import errorui


def exec_command(commands: list[str]):
    """Execute commands locally."""
    try:
        return check_output(commands).decode(sys.stdout.encoding)
    except CalledProcessError as error:
        errorui.show_dialog(error)


def get_tlp_config_file(prefix: str) -> str:
    """Select tlp config file by prefix."""
    return f"{prefix}/etc/tlp.conf"


def check_binaries_exist(flatpak_folder_prefix: str) -> None:
    """Check if required binaries are installed on system."""
    for expected_command in ["tlp", "tlp-stat", "lspci", "lsusb"]:
        if flatpak_folder_prefix != "":
            command_exists = Path(f"{flatpak_folder_prefix}/usr/bin/{expected_command}").exists()
            if not command_exists:
                command_exists = Path(f"{flatpak_folder_prefix}/usr/sbin/{expected_command}").exists()
        else:
            command_exists = which(expected_command) is not None
            if not command_exists:
                command_exists = Path(f"/usr/sbin/{expected_command}").exists()

        if not command_exists:
            errorui.show_dialog(f"{expected_command} not found on system. Please install first.")
            sys.exit(1)


def get_installed_tlp_version() -> str:
    """Fetch tlp version from command."""
    pattern = re.compile(r"TLP ([^\s]+)")
    current_config = exec_command(["tlp-stat", "-c"])
    matcher = pattern.search(current_config)
    return matcher.group(1)


def extract_tlp_base_version(raw_version: str) -> str:
    """Extract tlp base version from raw version."""
    pattern = re.compile(r"(\d\.\d+)")
    matcher = pattern.search(raw_version)
    return matcher.group(1).replace(".", "_")


def get_platform_profile_choices() -> str | None:
    """Fetch platform config choices from command."""
    allowed_choices = [
        'cool', 'quiet', 'low-power', 'balanced', 'balanced-performance', 'performance', 'custom', 'max-power'
    ]
    pattern = re.compile(r"/sys/firmware/acpi/platform_profile_choices\s+=\s+(.*)")
    current_platform_config = exec_command(["tlp-stat", "-p"])
    matcher = pattern.search(current_platform_config)
    if matcher:
        choices = matcher.group(1).strip().split(" ")
        if all(choice in allowed_choices for choice in choices):
            return ",".join(choices)
        else:
            print(f"Some platform profile choices are unknown.\n"
                  f"Discovered items {choices} not all in allowed {allowed_choices}.\n"
                  f"Please report at https://github.com/d4nj1/TLPUI/")
    return None


def get_user_config_file() -> Path:
    """Get config path for executing user."""
    userconfighome = getenv("XDG_CONFIG_HOME", "")
    if userconfighome == "":
        userconfigpath = Path(str(Path.home()) + "/.config/tlpui")
    else:
        userconfigpath = Path(str(userconfighome) + "/tlpui")
    return Path(str(userconfigpath) + "/tlpui.cfg")


class UserConfig:
    """Class to handle ui config parameters."""

    def __init__(self):
        """Init user config class parameters."""
        self.language = "en_EN"
        self.activeoption = 0
        self.activecategory = 0
        self.activeposition = 0
        self.windowxsize = 900
        self.windowysize = 600
        self.userconfigfile = get_user_config_file()
        self.read_user_config()

    def read_user_config(self):
        """Read ui config parameters from user home."""
        if self.userconfigfile.exists():
            config = configparser.ConfigParser()
            with open(str(self.userconfigfile), encoding='utf-8') as configfile:
                config.read_file(configfile)
            try:
                self.language = config['default']['language']
                self.activeoption = int(config['default']['activeoption'])
                self.activecategory = int(config['default']['activecategory'])
                self.activeposition = float(config['default']['activeposition'])
                self.windowxsize = int(config['default']['windowxsize'])
                self.windowysize = int(config['default']['windowysize'])
            except KeyError:
                # Config key error, override with default values
                self.write_user_config()
        else:
            self.userconfigfile.parent.mkdir(parents=True, exist_ok=True)
            self.write_user_config()

    def write_user_config(self):
        """Persist ui config parameters to user home."""
        config = configparser.ConfigParser()
        config['default'] = {}
        config['default']['language'] = self.language
        config['default']['activeoption'] = str(self.activeoption)
        config['default']['activecategory'] = str(self.activecategory)
        config['default']['activeposition'] = str(self.activeposition)
        config['default']['windowxsize'] = str(self.windowxsize)
        config['default']['windowysize'] = str(self.windowysize)
        with open(str(self.userconfigfile), mode='w', encoding='utf-8') as configfile:
            config.write(configfile)



"""Backend helper methods."""
import os


def detect_display():
    """
    Detect the appropriate GDK backend (wayland or x11).

    Prefers Wayland when the session is Wayland and the Wayland
    display socket is available. Otherwise falls back to x11.
    """
    session = os.environ.get("XDG_SESSION_TYPE", "").lower()

    # True Wayland session + available display socket
    if session == "wayland" and "WAYLAND_DISPLAY" in os.environ:
        return "wayland"

    # Fallback: if DISPLAY exists, use X11
    if "DISPLAY" in os.environ:
        return "x11"

    # If nothing else detected, default to x11
    print("Could not detect display backend. Falling back to x11.")
    return "x11"



"""This module provides general tlp-stat functions for the UI."""

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from . import language
from . import settingshelper
from .uihelper import get_theme_image, get_graphical_sudo
from .cmdhelper import create_sudo_command


def fetch_simple_stats(_, textbuffer: Gtk.TextBuffer) -> None:
    """Fetch simple tlp-stat information."""
    simple_stat_command = ["tlp-stat", "-r", "-t", "-c", "-s", "-u"]
    tlp_stat_output = settingshelper.exec_command(simple_stat_command)
    textbuffer.set_text(tlp_stat_output)


def fetch_complete_stats(_, textbuffer: Gtk.TextBuffer) -> None:
    """Fetch complete tlp-stat information."""
    sudo = get_graphical_sudo()

    if sudo is None:
        return

    sudo_cmd = create_sudo_command(sudo, ["tlp-stat"])
    tlp_stat_output = settingshelper.exec_command(sudo_cmd)
    textbuffer.set_text(tlp_stat_output)


def create_stat_box() -> Gtk.Box:
    """Create box with stat widgets."""
    scrolledwindow = Gtk.ScrolledWindow()
    scrolledwindow.set_hexpand(True)
    scrolledwindow.set_vexpand(True)

    textbuffer = Gtk.TextBuffer()
    textbuffer.set_text(language.ST_('Click fetch button to receive results'))

    textview = Gtk.TextView()
    textview.set_buffer(textbuffer)
    textview.set_editable(False)

    scrolledwindow.add(textview)

    emptylabel = Gtk.Label()

    fetchsimplebutton = Gtk.Button(label=f" {language.ST_('Simple')}",
                                   image=get_theme_image('dialog-information-symbolic', Gtk.IconSize.BUTTON))
    fetchsimplebutton.connect('clicked', fetch_simple_stats, textbuffer)
    fetchsimplebutton.set_always_show_image(True)

    fetchcompletebutton = Gtk.Button(label=f" {language.ST_('Complete')}",
                                     image=get_theme_image('format-indent-more-symbolic', Gtk.IconSize.BUTTON))
    fetchcompletebutton.connect('clicked', fetch_complete_stats, textbuffer)
    fetchcompletebutton.set_always_show_image(True)

    buttonbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)
    buttonbox.pack_start(emptylabel, True, True, 0)
    buttonbox.pack_start(fetchsimplebutton, False, False, 0)
    buttonbox.pack_start(fetchcompletebutton, False, False, 0)

    statbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=12)
    statbox.set_margin_top(18)
    statbox.set_margin_bottom(18)
    statbox.set_margin_left(18)
    statbox.set_margin_right(18)
    statbox.pack_start(buttonbox, False, False, 0)
    statbox.pack_start(scrolledwindow, True, True, 0)

    return statbox



"""Toggle config state."""

from gi.repository import Gtk
from .. import mainui
from .. import settings


def create_toggle_button(configname: str, configwidget: Gtk.Widget, window: Gtk.Window) -> Gtk.CheckButton:
    """Create state toggle."""
    togglebutton = Gtk.CheckButton()

    if settings.tlpconfig[configname].is_enabled():
        togglebutton.set_active(True)
    else:
        configwidget.set_sensitive(False)

    togglebutton.connect('toggled', on_button_toggled, configname, configwidget, window)
    return togglebutton


def on_button_toggled(self: Gtk.CheckButton, configname: str, configwidget: Gtk.Widget, window: Gtk.Window):
    """Process and store config state."""
    tlpobject = settings.tlpconfig[configname]

    if self.get_active():
        tlpobject.set_enabled(True)
        configwidget.set_sensitive(True)

        # Reset to default when intrinsic default gets reactivated
        if tlpobject.get_value() == "" and settings.tlpconfig_defaults[configname].is_enabled():
            tlpobject.set_value(settings.tlpconfig_defaults[configname].get_value())
            mainui.load_tlp_config(self, window, False)
    else:
        tlpobject.set_enabled(False)
        configwidget.set_sensitive(False)

        # If intrinsic default gets deactivated we have to remove value
        if settings.tlpconfig_defaults[configname].is_enabled():
            tlpobject.set_value("")



"""Provide translation."""

from . import settings
import gettext


def load_lang(langfile):
    """Load language from file."""
    translation = gettext.translation(langfile, settings.langdir, [settings.userconfig.language])

    versionlangfile = f"{langfile}{settings.tlpbaseversion}"
    if gettext.find(versionlangfile, settings.langdir, [settings.userconfig.language]) is None:
        return translation.gettext

    versiontranslation = gettext.translation(versionlangfile, settings.langdir, [settings.userconfig.language])
    versiontranslation.add_fallback(translation)
    return versiontranslation.gettext


CDT_ = load_lang('configdescriptions')
MT_ = load_lang('mainui')
ST_ = load_lang('statui')
UH_ = load_lang('uihelper')



"""Command helper functions."""


def create_sudo_command(sudo: str, command: list[str]) -> list[str]:
    """Create command for sudo."""
    if sudo == "pkexec":
        # if user selection required do it via UI or nothing
        sudo_cmd = ["pkexec", "--disable-internal-agent"]
    else:
        sudo_cmd = [sudo]

    return sudo_cmd + command



"""This module provides general helper functions for the UI."""

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GdkPixbuf
from shutil import which
from pathlib import Path
from . import language
from . import settings
from . import constants
from . import errorui


EXPECTED_ITEM_MISSING_TEXT = language.UH_('Expected item missing in config file')  # type: str
UNKNOWN_CONFIG_VALUE_TEXT = language.UH_('Unknown config value detected')  # type: str
SUDO_MISSING_TEXT = language.UH_('Install pkexec, gksu, gksudo, kdesu or kdesudo first')  # type: str
DEFAULT_STATE_TEXT = language.UH_('Default state:')  # type: str
DEFAULT_VALUE_TEXT = language.UH_('Default value:')  # type: str
CHANGED_STATE_TEXT = language.UH_('CHANGED')  # type: str


def get_graphical_sudo() -> str:
    """Fetch available graphical sudo command."""
    sudo_list = ["pkexec", "gksu", "gksudo", "kdesu", "kdesudo"]
    for sudo in sudo_list:
        if settings.IS_FLATPAK:
            sudo_exists = Path(f"{settings.FOLDER_PREFIX}/usr/bin/{sudo}").exists()
        else:
            sudo_exists = which(sudo) is not None

        if sudo_exists:
            return sudo
    errorui.show_dialog(SUDO_MISSING_TEXT)
    return None


def get_flag_image(locale: str) -> Gtk.Image:
    """Fetch flag image from icons folder."""
    flagpixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(f"{settings.icondir}flags/{locale}.png", width=16, height=16)
    return Gtk.Image().new_from_pixbuf(flagpixbuf)


def get_theme_image(iconname: str, iconsize: Gtk.IconSize) -> Gtk.Image:
    """Fetch image from theme or return fallback if missing."""
    if Gtk.IconTheme.get_default().has_icon(iconname):
        return Gtk.Image().new_from_icon_name(iconname, iconsize)
    return Gtk.Image().new_from_file(f"{settings.icondir}themeable/hicolor/scalable/actions/{iconname}.svg")


class StateImage:
    """Class to display configuration item state."""

    def __init__(self, defaultvalue, defaultstate, stateimage: Gtk.Image):
        """Init state image class parameters."""
        self.defaultvalue = str(defaultvalue)
        self.defaultstate = bool(defaultstate)
        self.stateimage = stateimage

    def warn_unknown_config_value(self, configvalue: str) -> None:
        """Add image and tooltip for unknown values."""
        self.stateimage.set_from_icon_name(constants.ICON_NAME_WARNING, Gtk.IconSize.BUTTON)
        self.stateimage.set_tooltip_text(f'{UNKNOWN_CONFIG_VALUE_TEXT}: {configvalue}')

    def refresh(self, value: str, store: str, enabled: bool, enabledstore: bool) -> None:
        """Refresh image and description by changed state."""
        changed = False
        if enabled != enabledstore or value != store:
            changed = True

        enabledtext = ''
        if enabled != self.defaultstate:
            enabledtext += f'{DEFAULT_STATE_TEXT} {str(self.defaultstate)}'

        if value == self.defaultvalue:
            if not changed and enabledtext == '':
                self.stateimage.clear()
            elif not changed and enabledtext != '':
                self.stateimage.set_from_icon_name(constants.ICON_NAME_INFO, Gtk.IconSize.BUTTON)
                self.stateimage.set_tooltip_text(f'{enabledtext}')
            elif changed and enabledtext == '':
                self.stateimage.set_from_icon_name(constants.ICON_NAME_UNDO, Gtk.IconSize.BUTTON)
                self.stateimage.set_tooltip_text(f'{CHANGED_STATE_TEXT}')
            elif changed and enabledtext != '':
                self.stateimage.set_from_icon_name(constants.ICON_NAME_UNDO, Gtk.IconSize.BUTTON)
                self.stateimage.set_tooltip_text(f'{CHANGED_STATE_TEXT}\n{enabledtext}')
        else:
            defaulttext = f'{DEFAULT_VALUE_TEXT} {self.defaultvalue}'
            if not changed and enabledtext == '':
                self.stateimage.set_from_icon_name(constants.ICON_NAME_INFO, Gtk.IconSize.BUTTON)
                self.stateimage.set_tooltip_text(defaulttext)
            elif not changed and enabledtext != '':
                self.stateimage.set_from_icon_name(constants.ICON_NAME_INFO, Gtk.IconSize.BUTTON)
                self.stateimage.set_tooltip_text(f'{enabledtext}\n{defaulttext}')
            elif changed and enabledtext == '':
                self.stateimage.set_from_icon_name(constants.ICON_NAME_UNDO, Gtk.IconSize.BUTTON)
                self.stateimage.set_tooltip_text(f'{CHANGED_STATE_TEXT}\n{defaulttext}')
            elif changed and enabledtext != '':
                self.stateimage.set_from_icon_name(constants.ICON_NAME_UNDO, Gtk.IconSize.BUTTON)
                self.stateimage.set_tooltip_text(f'{CHANGED_STATE_TEXT}\n{enabledtext}\n{defaulttext}')



"""Filehandling helper."""

import re
from io import open
from yaml import safe_load


def get_yaml_schema_object_from_file(objectname: str, filename: str) -> dict:
    """Read Yaml file."""
    with open(filename, encoding='utf-8') as yaml_file:
        yaml_object = safe_load(yaml_file)
    return yaml_object[objectname]


class TlpDefaults:
    """TLP defaults class."""

    def __init__(self, name: str, value: str, enabled: bool, quoted: bool):
        """Init TLP defaults class parameters."""
        self.name = name
        self.value = value
        self.enabled = enabled
        self.quoted = quoted

    def get_name(self) -> str:
        """Get defaults name."""
        return self.name

    def get_value(self) -> str:
        """Get defaults value."""
        return self.value

    def is_enabled(self) -> bool:
        """Get defaults enabled."""
        return self.enabled

    def is_quoted(self) -> bool:
        """Get defaults quoted."""
        return self.quoted


def extract_default_tlp_configs(filename: str) -> dict:
    """Fetch TLP defaults from file."""
    propertypattern = re.compile(r'^#?[A-Z_\d]+=')
    with open(filename, encoding='utf-8') as defaultsfile:
        lines = defaultsfile.readlines()

    tlpconfig_defaults = {}
    for line in lines:
        if propertypattern.match(line):
            cleanline = line.lstrip().rstrip()

            if cleanline.startswith('#'):
                enabled = False
                cleanline = cleanline.lstrip('#')
            else:
                enabled = True

            configproperty = cleanline.split('=', maxsplit=1)
            configname = configproperty[0]
            configvalue = configproperty[1]
            quoted = False

            if configvalue.startswith('\"') and configvalue.endswith('\"'):
                configvalue = configvalue.lstrip('\"').rstrip('\"')
                quoted = True

            tlpconfig_defaults[configname] = TlpDefaults(configname, configvalue, enabled, quoted)
    return tlpconfig_defaults
